Pro flux_calibrate,fluxfile,responsefile,wave=wave,response_wave=response_wave,$
                    mask=mask,float=float,calflux=calflux,nowrite=nowrite

;+---------------------------------------------------------------------------
;
; FLUX_CALIBRATE
;
; Perform relative flux calibration on a 1D or multi-order spectrum.
;
; INPUTS
;         fluxfile     - String, FITS filename of unclaibrated 1D
;                        spectrum.  Header must contain wavelength
;                        information, unless the 'wave' keyword is
;                        set.
;         resposnefile - String, FITS filename of response function, defined
;                        such that 
;                           calibrated flux = input flux * resposne fn
;                        Must have same number of orders (rows) as
;                        input flux.
;
; KEYWORDS 

;         wave         - String, FITS filename of wavelengths matching
;                        input 'fluxfile'.  Default is to use
;                        wavelength information in the 'fluxfile'
;                        header.
;         response_wave- String, FITS filename of wavelengths matching
;                        the input 'responsefile'.  Default is to use
;                        wavelength information in the 'responsefile' header,
;                        or to assume that the input flux and response
;                        functions cover the same wavelengths.
;         mask         - Scalar or 1D array of input flux pixels to mask.
;                        Masked pixels will be set to -1 on output.
;         float        - If set, forces conversion of output to 
;                        single-precision floating point.
;         calflux      - Array of calibrated flux values
;         nowrite      - If set, will not write output files.
;
; OUTPUTS
;         If input fluxfile = 'filename.fits', then calibrated flux array
;         will be written to 'filename_cal.fits'.
;
; HISTORY
;         Written 5/17/2005 GDB, based on HIRES_CALIBRATE
;         CALFLUX, NOWRITE keywords added 5/17/2007 GDB
;---------------------------------------------------------------------------

if (n_params() lt 2) then begin
   print,'CALLING SEQUENCE: '
   print,'   flux_calibrate,fluxfile,responsefile,wave=wave,'
   print,'                  response_wave=response_wave,mask=mask,float=float,'
   print,'                  calflux=calflux,/nowrite'
   return
endif

; Check that files exist
checkexists = findfile(fluxfile,count=nfiles)
if (nfiles eq 0) then begin
   print,'Flux file not found.'
   return
endif
checkexists = findfile(responsefile,count=nfiles)
if (nfiles eq 0) then begin
   print,'Response file not found.'
   return
endif

; Read in arrays
flux     = rflam(fluxfile,hflux,wflux)
response = rflam(responsefile,hresponse,wresponse)

; Promote response fn to floating point
response = 1.*response

; Read in wavelength arrays(optional)
if keyword_set(wave) then wflux = readfits(wave)
if keyword_set(response_wave) then begin
   wresponse = readfits(response_wave)
endif else begin
   if (total(wresponse) eq -1) then begin
      if (n_elements(flux) eq n_elements(response)) then begin
         wresponse = wflux
      endif else begin
         print,'FLUX-CALIBRATE: Must either provide response wavelength'
         print,'   information, or flux and response arrays must have the'
         print,'   same number of elements.'
         return
      endelse
   endif
endelse

; Get number of orders (rows) in input spectrum
sz = size(flux)
if (sz(0) eq 1) then n_orders=1 else n_orders = sz(2)

; Compute median non-zero wavelength in each row of the response
; function.
resp_sz = size(response)
if (resp_sz(0) eq 1) then n_resp_orders=1 else n_resp_orders = resp_sz(2)
resp_medwave = dblarr(n_resp_orders)
for i=0,n_resp_orders-1 do begin
   nonzero_ls = where(wresponse(*,i) gt 0)
   resp_medwave(i) = median(wresponse(nonzero_ls,i),/even)
endfor

; Flux calibrate each order.  Treat pixels with wavelenths < 0
; as missing data.

calflux = 0.*flux

for i=0,n_orders-1 do begin

   tmp_flux      = flux(*,i)
   tmp_wflux     = wflux(*,i)
   
   ; Find the response function order that best matches
   ; in wavelength
   nonzero_ls  = where(tmp_wflux gt 0)
   tmp_medwave = median(tmp_wflux(nonzero_ls),/even)
   diff        = abs(resp_medwave - tmp_medwave)
   dlam        = max(tmp_wflux(nonzero_ls)) - min(tmp_wflux(nonzero_ls))
   if (min(diff) gt 0.5*dlam) then begin
      print,'FLUX_CALIBRATE: No matching order in response function within'
      print,'                half an order.  Median wavelength = '+$
            strtrim(tmp_medwave,2)
stop
   endif
   k = where(diff eq min(diff))
   k = k(0)
   tmp_response  = response(*,k)
   tmp_wresponse = wresponse(*,k)
   
   ; Select pixels in the response function with wavelengths > 0
   pos_ls = where(tmp_wresponse gt 0)
   tmp_response  = tmp_response(pos_ls)
   tmp_wresponse = tmp_wresponse(pos_ls)
   
   ; Make sure response function is sorted in order of increasing
   ; wavelength
   sort_ls = sort(tmp_wresponse)
   tmp_response  = tmp_response(sort_ls)
   tmp_wresponse = tmp_wresponse(sort_ls)
   
   ; Interpolate response function onto flux wavelengths
   resp_match = interpol(tmp_response,tmp_wresponse,tmp_wflux)
   
   ; Compute calibrated flux
   calflux(*,i) = tmp_flux * resp_match

endfor

; Get root input filename (without extension)
trimfluxfile = strtrim(fluxfile,2)
len          = strlen(trimfluxfile)
lastdot      = strpos(trimfluxfile,'.',/reverse_search)
if (lastdot eq -1) then rootname = trimfluxfile else $
   rootname = strmid(trimfluxfile,0,lastdot)

; Preserve input data type if single-precision floating
sz   = size(flux)
ndim = sz(0)
type = sz(ndim+1)
if (type eq 4) then calflux = float(calflux)

; Force conversion to single-precision floating point (optional)
if keyword_set(float) then calflux = float(calflux)

; Write output file
if not(keyword_set(nowrite)) then writefits,rootname+'_cal.fits',calflux,hflux

return
end
