Pro lris_flatfixer,flatlist,wavearr,slitarr,combflat,combflatvar,$
                   range=range,illum=illum,bkspace=bkspace,$
                   pixflat=pixflat,fixstreaks=fixstreaks,blue=blue,red=red,$
                   newred=newred,hdf=hdf,gn=gn

;+-------------------------------------------------------------------
;
; LRIS_FLATFIXER    1/2007
;
; Median-combine and normalize LRIS longslit flats.
;
; Input flats are median combined, the combined flat is normalized,
; and the formal variance in the normalized flat is computed assuming
; Poisson statistics.
;
; Slit ilumination function is also divided out if the /fixstreaks
; keyword is set.
;
; INPUTS
;       flatlist  - String, name of single-column ASCII file containing
;                   the FITS filenames of flats to be combined (disabled),
;                   OR string array of FITS filenames of flats to be
;                   combined.
;       wavearr   - 2D array or fits filename, where the pixel values
;                   are uncalibrated wavelengths (in pixels).
;       slitarr   - 2D array or fits filename, where the pixel values
;                   are position along the slit.
;
; KEYWORDS
;       range      - Two-element array giving max and min slit
;                    positions of the illuminated area.  Regions
;                    outside this range will be filled with 1's in the
;                    normalized flat and -1's in the variance array.  Actual
;                    fitting is done in a slightly narrower window, to avoid
;                    edge effects.  Default is to automatically identify
;                    illuminated region.
;       illum      - Scalar, polymonial order+1 used to fit the slit
;                    illumination functin (1=constant, 2=linear, 3=
;                    quadrati, etc.).  Default is 1 (constant).  NOTE:
;                    Slit illumination function is taken out after
;                    bspline fitting is done in the dispersion
;                    direction, and not simultaneously at each break
;                    point.  If the pixflat keyword is set, then illum
;                    will be the order of the polynomial fit used for
;                    each column or row.  Default is 5 (5th order).
;       bkspace    - Spacing of bspline break points, in pixels (or
;                    whatever units are used for the input wavearr).
;                    Defaults set below are meant to fit out medium-scale
;                    features.  Smaller spacing may be needed for
;                    exposures that contain emission lines.  In
;                    general, however, a large spacing should be used
;                    so that fringe features are not divided out.
;       pixflat    - If set, will create a "pixel" flat by dividing
;                    out a high-order polynomial fit to the illumination
;                    function along each row (blue) or column (red).
;                    Default is to set a single low-order polynomial
;                    fit to the illumination function for all
;                    wavelengths.
;       fixstreaks - If set, will divide out small-scale residual
;                    illumination function (streaks), which includes
;                    variations in the slit illumination due to
;                    changes in the slit width, dust, etc.  Default is
;                    to take out large-scale illumination function
;                    that comes from the illumination of the slit by
;                    the internal flatfield lamps.
;       blue       - If set, assumes these are blue-side exposures.
;                    Default is to read instrument name from the 
;                    file header.
;       red        - If set, assumes these are red-side exposures.
;                    Default is to read instrument name from the 
;                    file header.
;       newred     - If setm assumes these are red-side exposures
;                    taken with then upgraded detector mosaic.
;                    Default is to read instrument name from the file
;                    header.
;       hdf        - If set, use settings to reduce custom 2" longslit
;                    flats used for HDF spectra.
;       gn         - Array of gain values, one per amplifier.
;                    Defaults are values for the original video
;                    boards.  Blue video board was replaced March
;                    2009.
;
;  OUTPUTS
;       combflat    - Median-combined, normalized flat field.  May
;                     either be a variable or a fits filename to which
;                     array is to be written.
;       combflatvar - Formal variance in the combined, normlized flat.
;                     May either be a variable or a fits filename to
;                     which array is to be written.
;
; HISTORY
;       Adapted from LDSS3_FLATFIXER  3/16/2008 GDB
;       PIXFLAT keyword added 3/8/2009 GDB
;       NEWRED keyword added 2/17/2012 GDB
;---------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:  '
   print,' lris_flatfixer,flatlist,wavearr,slitarr,combflat,combflatvar,'
   print,'                range=range,illum=illum,bkspace=bkspace,'
   print,'                /fixstreaks,/blue,/red,/newred,/hdf,gn=gn'
   return
endif

;;; Read in flat filenames
flatlist_sz = size(flatlist)
if (flatlist_sz(0) eq 0) then begin
   ; ASCII file with names of flat fields
   flat_templ = {   $
          version: 1.0, $
          datastart: 0L, $
          delimiter: ' ', $
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 1L, $
          fieldtypes: [7L], $
          fieldnames: ['name'], $
          fieldlocations: [6L], $
          fieldgroups: [0L] $
        }
   list     = read_ascii(flatlist,template=flat_templ)
   flatfile = list.name
endif else begin
   ; Array with names of flat fields
   flatfile = flatlist
endelse
nflat    = n_elements(flatfile)

;;; Blue or red exposures?
if (keyword_set(blue) or keyword_set(red) or keyword_set(newred)) then begin
   if keyword_set(blue) then begin
      blue   = 1
      red    = 0
      newred = 0
   endif
   if keyword_set(red) then begin
      blue   = 0
      red    = 1
      newred = 0
   endif
   if keyword_set(newred) then begin
      blue   = 0
      red    = 0
      newred = 1
   endif
endif else begin
   ; Read instrument name from header
   hdr = headfits(flatfile(0))
   instr = strtrim(sxpar(hdr,'INSTRUME'),2)
   naxis = sxpar(hdr,'NAXIS')
   case instr of
      'LRISBLUE': begin
                     blue   = 1
                     red    = 0
                     newred = 0
                  end
      'LRIS'    : begin
                     if (naxis eq 2) then begin
                        blue   = 0
                        red    = 1
                        newred = 0
                     endif else begin
                        blue   = 0
                        red    = 0
                        newred = 1
                     endelse
                  end
      else      : begin
                     print,'Instrument name not recognized'
                     return
                  end
   endcase
endelse

;;; Get binning
hdr     = headfits(flatfile(0))
binning = strtrim(sxpar(hdr,'BINNING'),2)
case binning of
   '1, 1': begin
              xbin = 1
              ybin = 1
           end
   ;'1, 2': begin
   ;           xbin = 1
   ;           ybin = 2
   ;        end
   ;'2, 1': begin
   ;           xbin = 2
   ;           ybin = 1
   ;        end
   '2,2':  begin
              xbin = 2
              ybin = 2
           end
   else:   begin
              print,'Error - Binning not recognized'
              return
           end
endcase

;;; Read in wave and slit grids.

if (size(wavearr,/tname) eq 'STRING') then begin
   checkexists = findfile(wavearr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Wavelength file not found.'
      return
   endif else begin
      wavegrid = readfits(wavearr)
   endelse
endif else wavegrid = wavearr

if (size(slitarr,/tname) eq 'STRING') then begin
   checkexists = findfile(slitarr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Slit position file not found.'
      return
   endif else begin
      slitgrid = readfits(slitarr)
   endelse
endif else slitgrid = slitarr

;;;
;;; Median combine flats
;;;

; Arrays to store processed flats.
sz         = size(wavegrid)
xsz        = sz(1)
ysz        = sz(2)
bigflat    = fltarr(xsz,ysz,nflat)
texp       = fltarr(nflat)         ; Exposure times

; Read in flats
for k=0,nflat-1 do begin
   tmp_flatfile = flatfile(k)
   checkexists  = findfile(tmp_flatfile,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Flat file '+tmp_flatfile+' not found.'
      return
   endif else begin
      bigflat(*,*,k) = lris_subbias(tmp_flatfile,flat_hdr,$
                                    blue=blue,red=red+newred,/flat)
      texp(k)        = sxpar(flat_hdr,'ELAPTIME')
   endelse
endfor

; Check that all flats have similar exposure times
texp_fracdiff = abs(texp - mean(texp))/mean(texp)
if (max(texp_fracdiff) gt 0.01) then begin
   print,'LRIS_FLATFIXER: Different exposure times detected.'
   return
endif

; Median combine
if (nflat gt 1) then begin
   medflat = median(bigflat,dim=3,/even)
endif else begin
   medflat = bigflat
endelse

;;;
;;; Normalize
;;;

; Take out differences in gain.  Allow for windowing in y (LRIS-R only)
if blue then begin
   if not(xbin eq 2) then begin
      print,'Only set up for LRIS-B xbin=2'
      return
   endif
   namp      = 4
   gain      = [1.542,1.546,1.632,1.633]
   amp_xmin  = [100,612,1124,1636]
   amp_xmax  = [611,1123,1635,2147]
endif
if red then begin
   if not(xbin eq 1) then begin
      print,'Only set up for LRIS-R xbin=1'
      return
   endif
   namp     = 2
   gain     = [1.98,2.13] ;2.19]
   amp_xmin = [40,1064]
   amp_xmax = [1063,2087]
endif
if newred then begin
   if not(xbin eq 2) then begin
      print,'Only set up for LRIS-R xbin=2 for new detector mosaic'
      return
   endif
   namp     = 4
   gain     = [1.255,1.180,1.191,1.162]
   amp_xmin = [1024,1536,0,512]
   amp_xmax = [1535,2047,511,1023]
endif
if keyword_set(gn) then gain = gn
; Equalize counts
mean_gain = mean(gain)
for i=0,namp-1 do $
   medflat(amp_xmin(i):amp_xmax(i),*) = $
      medflat(amp_xmin(i):amp_xmax(i),*) * (gain(i)/mean_gain)
; Use mean gain to calculate variance
mgn = mean_gain

; Choose read noise value close to the mean
if blue   then rdn = 4.0 ; e-
if red    then rdn = 6.2
if newred then rdn = 4.6  
  
; Set break point spacing for b-spline fit.
if keyword_set(bkspace) then bk_space = bkspace else begin
   if blue   then bk_space = 40. / ybin
   if red    then bk_space = 10. / xbin
   if newred then bk_space = 20. / ybin
endelse

; Identify min and max slit positions
xarr = lonarr(xsz,ysz)
yarr = lonarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
for i=0,ysz-1 do yarr(*,i) = i
if keyword_set(range) then begin
   slitmin = min(range)
   slitmax = max(range)
endif else begin
   ; Automatically identify illuminated region
   binsz    = 1.0
   h        = histogram(slitgrid,locations=slitbin,bin=binsz,rev=ri)
   nbin     = n_elements(h)
   flatbin  = fltarr(nbin)
   for i=0,nbin-1 do if (h(i) ne 0) then $
      flatbin(i) = median(medflat(ri(ri(i):ri(i+1)-1)))
   flatbin  = flatbin - min(flatbin)
   flatbin  = flatbin / max(flatbin)
   if blue then begin
      ; Avoid edges
      xlist = lindgen(xsz)
      illumbin = where(flatbin ge 0.3 and xlist ge 200 and xlist le xsz-200)
   endif
   if red then begin
      illumbin = where(flatbin ge 0.3)
   endif
   if newred then begin
      print,'Reducing LHS only!'
      xlist    = lindgen(xsz)
      illumbin = where(flatbin ge 0.3 and $
                       xlist ge 1475./xbin and xlist le 2020./xbin)
   endif
   slitmin  = min(slitbin(illumbin))
   slitmax  = max(slitbin(illumbin))+binsz
endelse
print,'Min slit = '+strtrim(slitmin,2)
print,'Max slit = '+strtrim(slitmax,2)

; Identify illumated region
if blue then begin
   ; Fit RHS and LHS separately.  Omit central 2 columns.
   ; Assumes 2x2 binning
   if (xbin ne 2) then begin
      print,'LRIS_FLATFIXER: Only set up for LRIS-B xbin=2'
   endif
   illum_ls1 = where(xarr ge 100 and xarr le 1122 and $
                     slitgrid ge slitmin and slitgrid le slitmax,n_illum)
   illum_ls2 = where(xarr ge 1125 and xarr le 2147 and $
                     slitgrid ge slitmin and slitgrid le slitmax,n_illum)
endif
if red then begin
   ; Avoid pre and overscan regions
   ; Assumes 1x1 binning
   if (xbin ne 1) then begin
      print,'LRIS_FLATFIXER: Only set up for LRIS-R xbin=1'
   endif
   illum_ls = where(xarr ge 40 and xarr le 2087 and $
                    slitgrid ge slitmin and slitgrid le slitmax,n_illum)
endif
if newred then begin
   illum_ls = where(slitgrid ge slitmin and slitgrid le slitmax,n_illum)
endif

; Choose narrow region to be fit in the dispersion direction.  Avoid edge
; regions.
edge_array = sobel(medflat)/float(medflat)
if blue then begin
   ; Fit RHS and LHS separately.  Omit central 2 columns.
   fit_ls1    = where(xarr ge 100 and xarr le 1122 and $
                      ;slitgrid ge slitmin and slitgrid le slitmax and $
                      slitgrid ge slitmin + 0.30*(slitmax-slitmin) and $
                      slitgrid le slitmin + 0.35*(slitmax-slitmin) and $
                      edge_array lt 1 and $
                      (shift(edge_array,[1,0]) lt 0.5 and $
                       shift(edge_array,[-1,0]) lt 0.5),n_fit)
   fit_ls2    = where(xarr ge 1125 and xarr le 2147 and $
                      ;slitgrid ge slitmin and slitgrid le slitmax and $
                      slitgrid ge slitmin + 0.60*(slitmax-slitmin) and $
                      slitgrid le slitmin + 0.65*(slitmax-slitmin) and $
                      edge_array lt 1 and $
                      (shift(edge_array,[1,0]) lt 0.5 and $
                       shift(edge_array,[-1,0]) lt 0.5),n_fit)
endif
if red then begin
   fit_ls     = where(xarr ge 40 and xarr le 2087 and $
                      ;slitgrid ge slitmin and slitgrid le slitmax and $
                      slitgrid ge slitmin + 0.65*(slitmax-slitmin) and $
                      slitgrid le slitmin + 0.70*(slitmax-slitmin) and $
                      edge_array lt 1 and $
                      (shift(edge_array,[0,1]) lt 0.5 and $
                       shift(edge_array,[0,-1]) lt 0.5),n_fit)
endif
if newred then begin
   fit_ls     = where(slitgrid ge slitmin + 0.60*(slitmax-slitmin) and $
                      slitgrid le slitmin + 0.70*(slitmax-slitmin) and $
                      edge_array lt 1 and $
                      (shift(edge_array,[1,0]) lt 0.5 and $
                       shift(edge_array,[-1,0]) lt 0.5),n_fit)
endif


; Fit a slowly-varying bspline in the dispersion direction.

if blue then begin
   fit_lhs = 1
   fit_rhs = 0
endif else begin
   fit_lhs = 0
   fit_rhs = 0
endelse

; Arrays to store normalize flat and variance
normflat     =  1 + fltarr(xsz,ysz)
normflat_var = -1 + fltarr(xsz,ysz)

done = 0

while not(done) do begin

   if blue then begin
      if fit_lhs then begin
         illum_ls = illum_ls1
         fit_ls   = fit_ls1
      endif else begin
         illum_ls = illum_ls2
         fit_ls   = fit_ls2
      endelse
   endif
   fitslit = slitgrid(fit_ls)
   fitwave = wavegrid(fit_ls)
   fitflat = medflat(fit_ls)
   ; Assume Poisson noise
   if (nflat gt 1) then med_factor = !pi/2. else med_factor = 1
   fitflat_var = med_factor*(fitflat/mgn + (rdn/mgn)^2)/nflat
   fitflat_invvar = 1./fitflat_var
   
   ; Identify potential bad pixels before fitting.
   maxwave = max(fitwave)
   minwave = min(fitwave)
   first_filter = 1
   filter_sz = 2*bk_space
   n_filters = ceil((maxwave-minwave)/filter_sz)
   ; Default is not to filter any points
   filter_ls = -1
   ; Speed things up by dividing arrays into regions that contain
   ; several filter bins.   *** Should rewrit this using HISTOGRAM ***
   h    = histogram(fitwave,bin=filter_sz,rev=ri)
   nbin = n_elements(h)
   for i=0,nbin-1 do begin
      if (h(i) ne 0) then begin
         tmp_ls     = ri(ri(i):ri(i+1)-1)
         tmp_flat   = fitflat(tmp_ls) 
         tmp_median = median(tmp_flat)
         tmp_sdev   = robust_sigma(tmp_flat)
         tmp_filter_ls = where(abs(tmp_flat-tmp_median) ge 5*tmp_sdev or $
                               tmp_flat lt tmp_sdev,n_tmp_filter)
         if (n_tmp_filter gt 0) then begin
            if first_filter then begin
               filter_ls    = tmp_ls(tmp_filter_ls)
               first_filter = 0 
            endif else begin
               filter_ls = [filter_ls,tmp_ls(tmp_filter_ls)]
            endelse
         endif
      endif
   endfor
   if (total(filter_ls) ne -1) then fitflat_invvar(filter_ls) = 0

   ; Perform b-spline fit (dispersion direction only)
   sset = bspline_iterfit(fitwave,fitflat,invvar=fitflat_invvar,$
                          bkspace=bk_space,yfit=fitflat_model,upper=5,lower=5)
   model_max = max(fitflat_model)
   
   ; Evaluate bspline fit over entire array and divide out
   medflat_model = fltarr(xsz,ysz)
   medflat_model(illum_ls) = bspline_valu(wavegrid(illum_ls),sset)
   normflat(illum_ls) = medflat(illum_ls) / medflat_model(illum_ls)
      
   ; Estimate variance in the normalized flat.  Note: this estimate will
   ; be wrong in the under-illuminated corners!  Non-illuminated regions 
   ; are flagged with -1.
   normflat_var(illum_ls) = $
         (med_factor*(medflat_model(illum_ls)/mgn + (rdn/mgn)^2)/nflat) / $
         (double(medflat_model(illum_ls))^2)

   ; Fit large-scale illumination function across the slit
   ;if keyword_set(hdf) then begin
   ;   ; Slit illumination correction for 2" custom longslit used for
   ;   ; HDF. Scaling counts ahead of fitting is done to remove
   ;   ; differences in counts due to varying thickness of the slit.
   ;   ; Assumes 2x2 binning for LRIS-B.  Choose relatively regions of
   ;   ; relatively smooth illumination and scale counts such that the
   ;   ; regions form a roughly continuous illumination function.
   ;   ; Scale values were determined empirically, but are close to
   ;   ; those determined from measuring the width of the slit in a
   ;   ; direct image.  Scale values are chosen to = 1 near the center.
   ;   slit_ls   = where(fitflat_model ge 0.5*model_max)
   ;   temp_slit = fitslit(slit_ls)
   ;   temp_norm = fitflat(slit_ls)/fitflat_model(slit_ls)
   ;   if blue then begin
   ;      if fit_lhs then begin
   ;         illum_order = 4 ; 3rd order
   ;         ls0 = where(temp_slit ge 466 and temp_slit le 706)
   ;         ls1 = where(temp_slit ge 743 and temp_slit le 1106)
   ;         scale = [1.085,1.000]
   ;         scaled_slit = [temp_slit(ls0),temp_slit(ls1)]
   ;         scaled_norm = [temp_norm(ls0)/scale(0),temp_norm(ls1)/scale(1)]
   ;      endif else begin
   ;         illum_order = 4 ; 3rd order
   ;         ls0 = where(temp_slit ge 1184 and temp_slit le 1254)
   ;         ls1 = where(temp_slit ge 1288 and temp_slit le 1454)
   ;         ls2 = where(temp_slit ge 1492 and temp_slit le 1679)
   ;         ls3 = where(temp_slit ge 1702 and temp_slit le 1967)
   ;         scale = [1.000,1.099,0.985,0.967]
   ;         scaled_slit = [temp_slit(ls0),temp_slit(ls1),$
   ;                        temp_slit(ls2),temp_slit(ls3)]
   ;         scaled_norm = [temp_norm(ls0)/scale(0),temp_norm(ls1)/scale(1),$
   ;                        temp_norm(ls2)/scale(2),temp_norm(ls3)/scale(3)]
   ;      endelse
   ;   endif else begin
   ;      illum_order = 6  ; 5th order
   ;      ls0 = where(temp_slit ge 119  and temp_slit le 451) 
   ;      ls1 = where(temp_slit ge 495  and temp_slit le 978)
   ;      ls2 = where(temp_slit ge 1049 and temp_slit le 1150)
   ;      ls3 = where(temp_slit ge 1192 and temp_slit le 1395)
   ;      ls4 = where(temp_slit ge 1468 and temp_slit le 1993)
   ;      ; These scale values are copied from the blue side.  The 
   ;      ; factor of 1.02 is the approximate change in slit width.
   ;      scale = [1.085,1.000,1.000*1.02,1.099*1.02,0.985*1.02]
   ;      scaled_slit = [temp_slit(ls0),temp_slit(ls1),$
   ;                     temp_slit(ls2),temp_slit(ls3),$
   ;                     temp_slit(ls4)]
   ;      scaled_norm = [temp_norm(ls0)/scale(0),temp_norm(ls1)/scale(1),$
   ;                     temp_norm(ls2)/scale(2),temp_norm(ls3)/scale(3),$
   ;                     temp_norm(ls4)/scale(4)]
   ;   endelse
   ;   pfit = poly_fit(scaled_slit,scaled_norm,illum_order-1,yfit=scaled_fit)
   ;   ; Evaluate across entire array and divide out
   ;   illum_model = fltarr(xsz,ysz)
   ;   for k=0,illum_order-1 do $
   ;      illum_model = illum_model + pfit(k)*(slitgrid^k)
   ;   normflat(illum_ls)     = normflat(illum_ls) / illum_model(illum_ls)
   ;   normflat_var(illum_ls) = normflat_var(illum_ls) / illum_model(illum_ls)
   ;   ; Take out rough residual difference in levels between rhs and
   ;   ; lhs ccds on the blue side.  Scale value measured empirically.
   ;   ; Note: This scaling is still probably not good enough to reduce
   ;   ; both ccds simultaneously.  Value may change if the flat field
   ;   ; lamp changes.
   ;   if blue and fit_rhs then begin
   ;      resid_scale = 1.08
   ;      normflat(illum_ls)     = normflat(illum_ls) * resid_scale
   ;      normflat_var(illum_ls) = normflat_var(illum_ls) * (resid_scale^2)
   ;   endif
   ;endif else begin


   ; Remove illumination function
   if keyword_set(pixflat) then begin
      ; Produce a pixel flat by fitting the illumination function
      ; at each row or column using a high-order polynomial
      if keyword_set(illum) then illum_order = illum else illum_order = 5
      illum_xy    = array_indices(medflat,illum_ls)
      illum_x     = illum_xy(0,*)
      illum_y     = illum_xy(1,*)
      illum_model = 1 + fltarr(xsz,ysz)
      if (blue or newred) then begin
         ; For each row, fit a polynomial over the min and max
         ; x-values of the illuminated region
         h = histogram(illum_y,bin=1,locations=y_list,rev=ri)         
         for j=0,n_elements(h)-1 do begin
            xmin  = min(illum_x(ri(ri(j):ri(j+1)-1)))
            xmax  = max(illum_x(ri(ri(j):ri(j+1)-1)))
            tmp_x = findgen(xmax-xmin+1)
            tmp_z = normflat(xmin:xmax,y_list(j))
            pfit  = robust_poly_fit(tmp_x,tmp_z,illum_order,tmp_zfit)
            illum_model(xmin:xmax,y_list(j)) = tmp_zfit
         endfor
      endif else begin
         ; For each column, fit a polynomial over the min and max
         ; y-values of the illuminated region
         h = histogram(illum_x,bin=1,locations=x_list,rev=ri)         
         for i=0,n_elements(h)-1 do begin
            ymin  = min(illum_y(ri(ri(i):ri(i+1)-1)))
            ymax  = max(illum_y(ri(ri(i):ri(i+1)-1)))
            tmp_y = findgen(ymax-ymin+1)
            tmp_z = normflat(x_list(i),ymin:ymax)
            pfit  = robust_poly_fit(tmp_y,tmp_z,illum_order,tmp_zfit)
            illum_model(x_list(i),ymin:ymax) = tmp_zfit
         endfor
      endelse
      normflat     = normflat / illum_model
      normflat_var = normflat_var / (illum_model^2)
   endif else begin
      ; Fit a single illumination function to the entire array.
      ; Default is to use a low-order polynomial.
      if keyword_set(illum) then illum_order = illum else illum_order = 1
      if (illum_order ge 2) then begin
         slit_ls = where(fitflat_model ge 0.5*model_max)
         pfit = poly_fit(fitslit(slit_ls),$
                         fitflat(slit_ls)/fitflat_model(slit_ls),$
                         illum_order-1,yfit=illum_fit)
         ; Evaluate across entire array and divide out
         illum_model = fltarr(xsz,ysz)
         for k=0,illum_order-1 do $
            illum_model = illum_model + pfit(k)*(slitgrid^k)
         normflat(illum_ls)     = normflat(illum_ls) / illum_model(illum_ls)
         normflat_var(illum_ls) = normflat_var(illum_ls) / (illum_model(illum_ls)^2)
      endif
   endelse

   ; Optionals - Remove remaining residual small-scale illumination 
   ; function (streaks)
   if keyword_set(fixstreaks) then begin
      if blue then begin
         resid_ls = where(medflat_model(illum_ls) ge 0.5*model_max)
      endif else begin
         wmin     = min(wavegrid(illum_ls))
         wmax     = max(wavegrid(illum_ls))
         dw       = wmax-wmin
         resid_ls = where(wavegrid(illum_ls) ge wmin+0.25*dw and $
                          wavegrid(illum_ls) le wmin+0.40*dw)
      endelse
      resid_slit = slitgrid(illum_ls(resid_ls))
      resid_flat = normflat(illum_ls(resid_ls))
      resid_var  = normflat_var(illum_ls(resid_ls))
      resid_sset = bspline_iterfit(resid_slit,resid_flat,invvar=1./resid_var,$
                                   bkspace=1.0,yfit=resid_fit,upper=5,lower=5)
      ; Evaluate residuals across entire array and divide out
      resid_model = fltarr(xsz,ysz)
      resid_model(illum_ls) = bspline_valu(slitgrid(illum_ls),resid_sset)
      normflat(illum_ls) = normflat(illum_ls) / resid_model(illum_ls)
      normflat_var(illum_ls) = normflat_var(illum_ls) / $
                               (resid_model(illum_ls)^2)
      ; For HDF data, flag region in the middle of the LRIS-R CCD
      ; covered by the slitmask bridge.
      if (keyword_set(hdf) and red) then begin
         bridge_indmin = min(where(resid_slit ge 900 and $
                                   resid_slit le 1100 and $
                                   resid_fit le 0.8))
         bridge_indmax = max(where(resid_slit ge 900 and $
                                   resid_slit le 1100 and $
                                   resid_fit le 0.8))
         bridge_slitmin = resid_slit(bridge_indmin)
         bridge_slitmax = resid_slit(bridge_indmax)
         bridge_ls = where(slitgrid ge bridge_slitmin-2 and $
                           slitgrid le bridge_slitmax+2)
         normflat(bridge_ls) = 1
         normflat_var(bridge_ls) = -1
      endif
   endif

   if blue and fit_lhs then begin
      fit_lhs = 0
      fit_rhs = 1
   endif else begin
      done = 1
   endelse

endwhile

; Restore differences in gain
for i=0,namp-1 do begin & $
   normflat(amp_xmin(i):amp_xmax(i),*) = $
      normflat(amp_xmin(i):amp_xmax(i),*) / (gain(i)/mean_gain) & $
   normflat_var(amp_xmin(i):amp_xmax(i),*) = $
      normflat_var(amp_xmin(i):amp_xmax(i),*) / ((gain(i)/mean_gain)^2) & $
endfor

;;;
;;; Set output arrays or write fits files
;;;

hnormflat = flat_hdr
sxaddhist,'-------------------------------------------------------',hnormflat
sxaddhist,'Flatfield processing done using LRIS_FLATFIXER',hnormflat
sxaddhist,'   Flatfield frames combined:',hnormflat
for i=0,nflat-1 do $
   sxaddhist,'      '+flatfile(i),hnormflat
sxaddhist,'-------------------------------------------------------',hnormflat

sxaddpar,hnormflat,'BZERO',0.0

if (size(combflat,/tname) eq 'STRING') then begin
   print,'Writing '+combflat
   writefits,combflat,normflat,hnormflat
endif else combflat = normflat

if (size(combflatvar,/tname) eq 'STRING') then begin
   print,'Writing '+combflatvar
   writefits,combflatvar,normflat_var,hnormflat
endif else combflatvar = normflat_var

print,'Done.'

return
end
