Pro lris_slitgrid,array,slitgrid,blue=blue,red=red,newred=newred,edges=edges,$
                  objects=objects,disprange=disprange

;+-------------------------------------------------------------------------
;
; LRIS_SLITGRID         9/2006
;
; Create 2D array of slit positions.  The slit position at a pixel is
; the x- (y)-intercept of the curve of constant wavelength that would
; pass through the center of that pixel for the blue (red) side.  For
; the blue side, coordinates are taken w/r/t the left-hand chip.
; Therefore, there will be a jump between the slit positions on the
; left and right-hand chips to account for the chip offset and slight
; rotation.
;
; Default is to trace streaks in flat field.  May also trace flatfield
; edges by setting the /edges keyword, or trace multiple objects by
; setting the /objects keyword.
; 
; Must set either the /blue or /red keyword.
;
; INPUTS
;      array        - String, FITS filename of array to trace,
;                     flat if tracing streaks, or object exposure
;                     if the objects keyword is set.  If an array
;                     of filenames is given, then the exposures
;                     will be added before tracing.
;
; KEYWORDS
;      blue         - Blue side
;      red          - Red side
;      newred       - Red side with upgraded detector mosaic
;      edges        - Trace slit edges instead of streaks
;      objects      - Trace objects instead of streaks
;      disprange    - Two element array, [min,max] values for
;                     display scaling.
;
; OUTPUTS
;      slitgrid     - 2D array or name of FITS file, slit positions
;                     for each pixel.
;
; HISTORY
;      Adopted from NIRSPEC_SLITGRID 9/13/2006 GDB
;      NEWRED keyword added 2/17/2012 GDB
;      Allow multiple input exposures 2/17/2012 GDB
;--------------------------------------------------------------------------

if (n_params() lt 2) then begin
 print,'CALLING SEQUENCE: '
 print,'   lris_slitgrid,array,slitgrid,/blue,/red,/newred,edges=edges,'
 print,'                 disprange=disprange'
 return
endif

;;; Is this for the blue side or the red side?
if not(keyword_set(blue) or keyword_set(red) or keyword_set(newred)) $
   then begin
   print,'Must set either /blue or /red or /newred keywords.'
   return
endif
if keyword_set(blue) then begin
   arm    = 'blue'
   blue   = 1
   red    = 0
   newred = 0
   ydir   = 1
   xdir   = 0
endif
if keyword_set(red) then begin
   arm    = 'red'
   blue   = 0
   red    = 1
   newred = 0
   ydir   = 0
   xdir   = 1
endif
if keyword_set(newred) then begin
   arm    = 'newred'
   blue   = 0
   red    = 0
   newred = 1
   ydir   = 1
   xdir   = 0
endif

;;; Read in array to trace
if (n_elements(array) gt 1) then begin
   ; Add multiple exposures
   tmp_arr = lris_subbias(array(0),harr,blue=blue,red=red+newred)
   sz   = size(tmp_arr)
   xsz  = sz(1)
   ysz  = sz(2)
   big_arr = fltarr(xsz,ysz,n_elements(array))
   big_arr(*,*,0) = tmp_arr
   for i=1,n_elements(array)-1 do $
      big_arr(*,*,i) = lris_subbias(array(i),harr,blue=blue,red=red+newred)
   arr = total(big_arr,3)
endif else begin
   ; Single exposure
   arr  = lris_subbias(array,harr,blue=blue,red=red+newred)
endelse
sz   = size(arr)
xsz  = sz(1)
ysz  = sz(2)
xcenter = round(xsz/2.)
ycenter = round(ysz/2.)

;;; Binning?
binning = strtrim(sxpar(harr,'BINNING'),2)
case binning of
   '1, 1': begin
              xbin = 1
              ybin = 1
           end
   ;'1, 2': begin
   ;           xbin = 1
   ;           ybin = 2
   ;        end
   ;'2, 1': begin
   ;           xbin = 2
   ;           ybin = 1
   ;        end
   '2,2':  begin
              xbin = 2
              ybin = 2
           end
   else:   begin
              print,'Error - Binning not recognized'
              return
           end
endcase

;;; Set trace range
case arm of
   'blue'   : trace_range = [0,0.83*ysz]   ; Faint in the far blue
   'red'    : trace_range = [40,2087]      ; Red assumes 1x1 binning
   'newred' : trace_range = [0,ysz-1]
endcase

;;; Display a region of the arr and allow the user to identify
;;; streaks.  Shrink section to fit on screen, if necessary.
scrsz  = get_screen_size()
xscrsz = scrsz(0)
yscrsz = scrsz(1)
xfactor = (xsz / (0.8*xscrsz)) > 1.0
yfactor = (ysz / (0.8*yscrsz)) > 1.0
if (blue or newred) then begin
   scalefactor = xfactor
   ;xside = (xsz/4.) < (xscrsz/2.-50)
   ;xmin = xcenter - xside
   ;xmax = xcenter + xside
   xmin = 0.
   xmax = floor(xsz / scalefactor) - 1
   ymin = ceil((ycenter - 100.) / scalefactor)
   ymax = floor((ycenter + 100.) / scalefactor)
endif else begin
   print,'Assuming red binning is 1x1.'
   scalefactor = yfactor
   ;yside = 490. < (yscrsz/2.-50)
   xmin = ceil((xcenter - 100.) / scalefactor)
   xmax = floor((xcenter + 100.) / scalefactor)
   ;ymin = ycenter - yside
   ;ymax = ycenter + yside
   ymin = 0.
   ymax = floor(ysz / scalefactor) - 1
endelse
disp_arr = congrid(arr,xsz/scalefactor,ysz/scalefactor)
medval  = median(disp_arr(xmin:xmax,ymin:ymax))
chan,9
if keyword_set(edges) then begin
   if keyword_set(disprange) then begin
      doit,disp_arr(xmin:xmax,ymin:ymax),disprange(0),disprange(1)
   endif else begin
      doit,disp_arr(xmin:xmax,ymin:ymax),0.8*medval,2*medval
   endelse
endif else begin
   if keyword_set(disprange) then begin
      doit,disp_arr(xmin:xmax,ymin:ymax),disprange(0),disprange(1)
   endif else begin
      doit,disp_arr(xmin:xmax,ymin:ymax),0.9*medval,1.1*medval
   endelse
endelse
if keyword_set(edges) then begin
   print,'Left-click on edges to trace.  Right-click when done.'
endif else begin
   if keyword_set(objects) then begin
      print,'Left-click on objects to trace.  Right-click when done.'
   endif else begin
      print,'Left-click on streaks to trace.  Right-click when done.'
   endelse
endelse
getpoints,x0_raw,y0_raw,/device,/down
x0 = scalefactor * (x0_raw + xmin)
y0 = scalefactor * (y0_raw + ymin)

;;; Trace each line and fit the traced points
trace_order = 2
nlines      = n_elements(x0)
a           = fltarr(trace_order+1,nlines)
aerr        = fltarr(trace_order+1,nlines)
chan,11
ntiles2 = floor(sqrt(nlines))
ntiles1 = ceil((1.*nlines)/ntiles2)
!p.multi=[0,ntiles1,ntiles2]
for i=0,nlines-1 do begin
   chan,11
   ; Default is to trace a negative of the arr in order to fit
   ; Guassians to streaks.  Alternatively, trace edges of flat field
   ; or objects.
   if keyword_set(edges) then begin
      trace_arr = arr
      by_gauss  = 0
      sub_pix   = 10
   endif else begin
      if keyword_set(objects) then begin
         trace_arr = arr
         by_gauss  = 1
         sub_pix   = 0
      endif else begin
         trace_arr = -arr
         by_gauss  = 1
         sub_pix   = 0
      endelse
   endelse
   traceline,trace_arr,x0(i),y0(i),xtr,ytr,xdir=xdir,ydir=ydir,nside=10,$
             median=7,merit=merit,/noplot,range=trace_range,$
             bygauss=by_gauss,subpix=sub_pix
   if blue then begin
      ; Adjust x and y values of pixels on the RH chip to account
      ; for the chip offset and rotation.
      xoffset = 103.192 / xbin
      yoffset = -2.982 / ybin
      xscale  = 0.00162 * (ybin/xbin)
      ; Assume pre-scan region is present
      case xbin of
        1: rh_ls = where(xtr ge 2252,n_rh)
        2: rh_ls = where(xtr ge 1124,n_rh)      
        else : begin
                   print,'Error - Not sure how to handle binning.'
                   return
               end
      endcase
      if (n_rh ne 0) then begin
         xtr(rh_ls) = xtr(rh_ls) + xoffset + xscale*(ytr(rh_ls)+1)
         ytr(rh_ls) = ytr(rh_ls) + yoffset
      endif
      pixtr = ytr    ; Independent variable
      ztr   = xtr    ; Dependent variable
   endif else begin
      if red then begin
         pixtr = xtr
         ztr   = ytr
      endif else begin
         pixtr = ytr
         ztr   = xtr
      endelse
   endelse
   ; Fit trace points.  Reject outliers and refit until fit converges.
   fit_converged = 0
   old_zfit      = 0
   n_iter        = 0
   good_ls       = where(merit ne 0)
   while not(fit_converged) do begin
      temp_a   = poly_fit(pixtr(good_ls),ztr(good_ls),trace_order,$
                          yerror=zerr,sigma=temp_aerr)
      zfit = 0.*pixtr
      for j=0,trace_order do zfit = zfit + temp_a(j)*pixtr^j
      zfit_diff = zfit - old_zfit
      if (max(abs(zfit_diff)) eq 0) then begin
         fit_converged = 1
      endif else begin
         good_ls = where(abs(ztr-zfit) le 4*zerr and merit ne 0,n_good)
         if (n_good lt trace_order+1) then begin
            ; Flag bad fit
            print,'Bad fit - rejecting.'
            temp_a(*)     = -9999
            temp_aerr(*)  = -9999
            fit_converged = 1
         endif
      endelse
      n_iter = n_iter + 1
      if (n_iter ge 10) then fit_converged = 1
   endwhile
   a(*,i)    = temp_a
   aerr(*,i) = temp_aerr
   ;; Plot fit residuals
   ;zresid = ztr - zfit
   ;zresid_sdev = robust_sigma(zresid)
   ;chan,12
   ;plot,pixtr,zresid,/xstyle,/ystyle,psym=3,title='Trace residuals',$
   ;   xtitle='x',ytitle='y',xrange=[-5*zresid_sdev,5*zresid_sdev]
   ; Plot trace with fit
   plot,xtr,ytr,/xstyle,/ystyle,psym=3,xtitle='x',ytitle='y'
   if (blue or newred) then oplot,zfit,pixtr else oplot,pixtr,zfit
endfor

;;; Fit coefficients as a function of x (or y)-intercept.
p  = fltarr(trace_order,4)
if (nlines gt 1) then begin
   ; Reject bad fits
   fitls = where(a(0,*) ne -9999 and a(0,*) ge -500 and a(0,*) le 2000)
   a_0 = a(0,fitls)
   chan,10
   !p.multi = [0,1,1]
   for i=0,trace_order-1 do begin
      print,' '
      print,'Fitting term '+strtrim(i+1,2)
      a_i    = a(i+1,fitls)
      aerr_i = aerr(i+1,fitls)
      p_i    = inter_poly_fit(a_0,a_i,1,include=inc,measure_errors=aerr_i)
      p(i,0:n_elements(p_i)-1) = p_i
   endfor
endif else begin
   p(*,0) = a(1:trace_order)
endelse

;;; Construct the slit position grid.
print,'Computing slit grid...'
if blue then begin
   ;case xbin of
   ;   1: left_xsz = 2252
   ;   2: left_xsz = 1124
   ;   else: begin
   ;            print,'Error - Not sure how to handle binning.'
   ;            return
   ;         end
   ;endcase
   ;right_xsz = xsz - left_xsz
   ;left_x = lindgen(left_xsz)
   ;left_y = lindgen(ysz)
   ;raw_right_x = lindgen(right_xsz) + left_xsz
   ;raw_right_y = lindgen(ysz)
   ;right_x = raw_right_x + xoffset + xscale*(raw_right_y+1)
   ;right_y = raw_right_y + yoffset
   ;left_slitarr = xy2a(left_x,left_y,p)
   ;right_slitarr = xy2a(right_x,right_y,p)   
   ;slitarr = fltarr(xsz,ysz)
   ;slitarr(0:left_xsz-1,*)   = left_slitarr
   ;slitarr(left_xsz:xsz-1,*) = right_slitarr
   x = indgen(xsz)
   y = indgen(ysz)   
   slitarr = float(xy2a(x,y,p,/lrisb,xbin=xbin,ybin=ybin))
endif
if red then begin
   x = indgen(xsz)
   y = indgen(ysz)
   slitarr = float(xy2b(x,y,p))
endif
if newred then begin
   x = indgen(xsz)
   y = indgen(ysz)
   slitarr = float(xy2a(x,y,p))
endif

;;; Optional - write to file.
if (size(slitgrid,/tname) eq 'STRING') then begin
   sxaddpar,harr,'BZERO',0.0
   writefits,slitgrid,slitarr,harr
endif else begin
   slitgrid = slitarr
endelse

print,'Done.'

return
end
