Pro lris_wavecalib,arcfile,slitgrid,wave_uncalib,wave_calib,$
                 linelist=linelist,dispersion=dispersion,idlist=idlist,$
                 cutcent=cutcent,_extra=_extra

;+----------------------------------------------------------------------------
;
; LRIS_WAVECALIB
;
; Wavelength calibrate LRIS longslit orders by hand.
;
; INPUTS
;         arcfile      - FITS filename of arc lamp exposure
;         slitgrid     - FITS filename of slit position map (see
;                        lris_slitgrid.pro)
;         wave_uncalib - FITS filename of ncalibrated wavelength map,
;                        as produced by lris_wavecalib.
;
; KEYWORDS
;         linelist     - String, name of ASCII file containing line
;                        wavelengths.  Default is HgNeArCdZn lamp
;                        list suitable for low resolution.  Wavelengths
;                        should be listed in the first column and have
;                        units of angstroms.
;         dispersion   - Approximate dispersion, in units of
;                        units(linelist)/units(wave_uncalib).  This is
;                        generally angstroms/pixel.  Default is 1.175
;                        (VPH Red grism).  However, if using skylines
;                        to tweak an already-calibrated array, the
;                        units are angstroms/ansgtrom, and dispersion
;                        should be set to 1.
;         idlist       - String, name of 2-column ASCII file to which to
;                        write x-positions and wavelengths of identified
;                        lines.  Format is row = {x wavelength}.
;         cutcent      - Scalar, x value of column (blue) or y value
;                        of row (red) around which to take a cut through
;                        the array when identifyng lines.  Default
;                        is a point near the center.
;         _extra       - WAVE_CALIB keywords, esp. 'guess'
;
; OUTPUTS
;         wave_calib   - Name of file to which to write calibrated
;                        wave grid as a multi-extension FITS file.
;
; HISTORY
;         Adapted from LDSS3_WAVECALIB  3/13/2008 GDB
;         CUTCENT keyword added 3/7/2009 GDB
;-----------------------------------------------------------------------------

if (n_params() lt 4) then begin
   print,'CALLING SEQUENCE:'
   print,'   lris_wavecalib,arcfile,slitgrid,wave_uncalib,wave_calib,'
   print,'                   linelist=linelist,dispersion=dispersion,'
   print,'                   idlist=idlist,cutcent=cutcent,_extra=_extra'
   return
endif

; Read in slit grid and uncalibrated wavelength grid
slit    = readfits(slitgrid,hslit)
uncalib = readfits(wave_uncalib,huncalib)

; Read in arc and subtract bias
harc  = headfits(arcfile)
instr = strtrim(sxpar(harc,'INSTRUME'),2)
naxis = sxpar(harc,'NAXIS')
case instr of
   'LRISBLUE': begin
                  arc    = lris_subbias(arcfile,/blue)
                  blue   = 1
                  red    = 0
                  newred = 0
               end
   'LRIS'    : begin
                  if (naxis eq 2) then begin
                     arc    = lris_subbias(arcfile,/red)
                     blue   = 0
                     red    = 1
                     newred = 0
                  endif else begin
                     arc    = lris_subbias(arcfile,/red)
                     blue   = 0
                     red    = 0
                     newred = 1
                  endelse
               end
   else      : begin
                  print,'Instrument name not recognized'
                  return
               end
endcase

; Calibrated wavelength arrays
calib = 0d0*uncalib

; Select pixels just for wavelength calibration.  For LRIS, the
; the illumination changes rapidly across the slit, and in
; different ways for different lines (different lamps?).  Therefore,
; easiest to just take a median cut through the array, rather than
; attempt to subsample the lines. 
sz   = size(arc)
xsz  = sz(1)
ysz  = sz(2)
if blue then begin
   xmid = 0.52*xsz
   if keyword_set(cutcent) then xmid = cutcent
   uncalib_cut = median(uncalib(xmid-2:xmid+2,*),dim=1)
   arc_cut     = median(arc(xmid-2:xmid+2,*),dim=1)
endif
if red then begin
   ymid = 0.52*ysz
   if keyword_set(cutcent) then ymid = cutcent
   uncalib_cut = median(uncalib(*,ymid-2:ymid+2),dim=2)
   arc_cut     = median(arc(*,ymid-2:ymid+2),dim=2)
endif
if newred then begin
   xmid = 0.43*xsz
   if keyword_set(cutcent) then xmid = cutcent
   uncalib_cut = median(uncalib(xmid-2:xmid+2,*),dim=1)
   arc_cut     = median(arc(xmid-2:xmid+2,*),dim=1)
endif


; Get wavelength calibration coefficients
if keyword_set(linelist) then line_list = linelist else $
   line_list = '~/calib/lris/linelists/line_list_air_ordered.dat'
   ; Air wavelengths!
coeffs = 0
xid    = 0
waveid = 0
; Dispersion, in units of calibrated wavelength/uncalibrated wavelength
if keyword_set(dispersion) then  disp=dispersion else disp=1.175
; Rough resolution (angstroms)
;resoln = 4.7  ; 600/4000 Grism w/ 1" slit
resoln = 4*disp
wave_calib,uncalib_cut,arc_cut,linefile=line_list,$
           coeffs=coeffs,xid=xid,waveid=waveid,dispersion=disp,$
           resolution=resoln,psym=10,_extra=_extra
n_id = n_elements(waveid)
; Calculate calibrated wavelength values.  If only one line was
; identified, then perform a bulk shift.
poly_order = n_elements(coeffs) - 1
calib      = 0d0*uncalib
if (n_id eq 1) then begin
   calib = uncalib + waveid(0) - xid(0)
endif else begin 
   for h=0,poly_order do calib = calib + coeffs(h)*(uncalib^h)
endelse

; Write array(s) to FITS file.
writefits,wave_calib,calib,huncalib

; Optional - write identified lines to a file
if keyword_set(idlist) then begin
   openw,unit,idlist,/get_lun
   for i=0,n_id-1 do printf,unit,xid(i),waveid(i)
   close,unit
endif

return
end

