Pro lris_wavegrid,exposure,slitgrid,wavegrid,slitrange=slitrange,$
                  list=list,outlist=outlist,blue=blue,red=red,newred=newred,$
                  mask=mask,nside=nside,cutcent=cutcent,manual=manual,$
                  sky=sky

;+-------------------------------------------------------------------------
;
; LRIS_WAVEGRID         4/2005
;
;
; Trace skylines or arclines in a 2D spectrum and create a 2D array of
; uncalibrated wavelengths.  The uncalibrated wavelength at a pixel is
; the y- (x-)intercept of the curve of constant wavelength that would
; pass through the center of that pixel for the blue (red) side.  For
; the blue side, coordinates are taken w/r/t the left-hand chip.
; Therefore, there will be a jump between the wave positions on the
; left and right-hand chips to account for the chip offset and slight
; rotation.
;
; Must set either the /blue or /red or /newred keywords.
;
; The user may either select lines by hand or, optionally,
; identify lines in an input file (see below).
;
; *** NOTE: Somehow this method doesn't work as smoothly as it does
;           for other instruments.  Some lines just seem to be
;           outliers, and the user has to carefully select which
;           lines to include during the interactive step.  At least
;           on the red side, the outliers sometimes form a sequence
;           of their own.  These must either be relected lines, or
;           else the different arc lamps illuminate the slit
;           slightly differently.
;
; INPUTS
;      exposure    - 2D array or FITS filename, raw object frame with
;                    pronounced skylines, or arc lamp exposure.
;      slitgrid    - 2D array or FITS filename, frame with slit
;                    position at each pixel.  Typically, the slit
;                    position is the x-intercept of a line of constant
;                    slit position that passes through the center of
;                    that pixel.
;      
; KEYWORDS
;      slitrange   - 2-element array, minimum and maximum slit
;                    positions over which to trace sky or arc lines.
;                    Defaults covers a typical low-resolution LRIS
;                    exposures.
;      list        - String, name of 2-column ascii file containing x
;                    and y positions of lines to trace.  Default is
;                    to allow the user to select lines from a cut
;                    across the input exposure.
;      outlist     - String, name of 2-column ascii file to which to
;                    write the x and y positions of the identified
;                    lines.  If the keyword is set without any value,
;                    line positions will be written to 'list.dat'.
;      blue        - Blue side
;      red         - Red side
;      newred      - Red side with upgraded detector mosaic
;      mask        - 2-element array, minimum and maximum slit
;                    values to mask out when tracing lines.  Useful
;                    if tracing lines from a mask with a bridge.
;      nside       - Number of pixels to either side of a line to
;                    use when computing centroid.  Default is 7,
;                    which is okay for a 1" slit.  Wider slits might
;                    require larger values.
;      cutcent     - Scalar, x (if blue) or y (if red) value at which
;                    to take a cut through the array to identify lines.
;                    Useful when setting the slitrange keyword.  Default
;                    is to select a value on the right hand (blue) or
;                    near the center (red) of the array.
;      manual      - If set, manually identify lines to trace.  Default
;                    is to select lines automatically for the red side.
;      sky         - Use cross-correlation techniques to trace sky
;                    features.  Useful for creating a wave grid from
;                    a blue sky exposure, which has few well-defined
;                    lines.  Default is to trace individual lines by
;                    fitting Gaussians.
;
; OUTPUTS
;      wavegrid    - 2D array or FITS filename, uncalibrated wavelengths.
;                    If given as a string, the output will be written
;                    to a file with that name.
;
; HISTORY
;      Adopted from NIRSPEC_WAVEGRID 9/14/2006 GDB
;      CUTCENT, MANUAL keywords added 3/7/2009 GDB
;      SKY keyword added 6/4/2009 GDB
;      NEWRED keyword added 2/17/2012 GDB
;--------------------------------------------------------------------------

if (n_params() lt 3) then begin
 print,'CALLING SEQUENCE: '
 print,'   lris_wavegrid,exposure,slitgrid,wavegrid,slitrange=slitrange,'
 print,'                 list=list,outlist=outlist,/blue,/red,/newred,'
 print,'                 mask=mask,nside=nside,cutcent=cutcent,/manual,/sky' 
 return
endif

;;; Is this for the blue side or the red side?
if not(keyword_set(blue) or keyword_set(red) or keyword_set(newred)) $
   then begin
   print,'Must set either /blue or /red or /newred keywords.'
   return
endif
if keyword_set(blue) then begin
   arm    = 'blue'
   blue   = 1
   red    = 0
   newred = 0
   ydir   = 0
   xdir   = 1
endif
if keyword_set(red) then begin
   arm    = 'red'
   blue   = 0
   red    = 1
   newred = 0
   ydir   = 1
   xdir   = 0
endif
if keyword_set(newred) then begin
   arm    = 'newred'
   blue   = 0
   red    = 0
   newred = 1
   ydir   = 0
   xdir   = 1
endif

;;; Read in arrays, if necessary.

if (size(exposure,/tname) eq 'STRING') then begin
   checkexists = findfile(exposure,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Exposure file not found.'
      return
   endif else begin
      obj = lris_subbias(exposure,hobj,blue=blue,red=red+newred)
      ;;; Binning?
      binning = strtrim(sxpar(hobj,'BINNING'),2)
      case binning of
         '1, 1': begin
                    xbin = 1
                    ybin = 1
                 end
         ;'1, 2': begin
         ;           xbin = 1
         ;           ybin = 2
         ;        end
         ;'2, 1': begin
         ;           xbin = 2
         ;           ybin = 1
         ;        end
         '2,2':  begin
                    xbin = 2
                    ybin = 2
                 end
         else:   begin
                    print,'Error - Binning not recognized'
                    return
                 end
      endcase
   endelse
endif else begin
   obj = exposure
   print,'No input file.  Assuming 1x1 binning.'
   xbin = 1
   ybin = 1
endelse

if (size(slitgrid,/tname) eq 'STRING') then begin
   checkexists = findfile(slitgrid,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Slitgrid file not found.'
      return
   endif else begin
      slit = readfits(slitgrid,hslit)
   endelse
endif else slit = slitgrid

sz  = size(obj)
xsz = sz(1)
ysz = sz(2)

; Screen size
device,get_screen_size=ssz

;;; Take a cut along the sky and identify lines, or read in line
;;; positions from a file.
if keyword_set(list) then begin
   templ =  {   $
          version: 1.0, $
          datastart: 0L, $
          delimiter: ' ', $
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 2L, $
          fieldtypes: [4L,4L], $
          fieldnames: ['x','y'], $
          fieldlocations: [6L,23L], $
          fieldgroups: [0L,1L] $
        }
   position = read_ascii(list,template=templ)
   linex    = position.x
   liney    = position.y
   nlines   = n_elements(linex)
endif else begin
   if blue then begin
      if keyword_set(cutcent) then begin
         xcut_cent = cutcent
      endif else begin
         case xbin of
            1: xcut_cent = 2275
            2: xcut_cent = 1407  ; Avoid refelcted arc lines
            else: begin
                     print,'Error - Binning not recognized'
                     return
                  end
         endcase
      endelse
      cut = median(obj(xcut_cent-5:xcut_cent+5,*),dim=1)
      z   = lindgen(ysz)
   endif
   if red then begin
      if keyword_set(cutcent) then begin
         ycut_cent = cutcent
      endif else begin
         ycut_cent = 1.2*round(ysz/2.)
      endelse
      cut = median(obj(*,ycut_cent-5:ycut_cent+5),dim=2)
      z   = lindgen(xsz)
   endif
   if newred then begin
      if keyword_set(cutcent) then begin
         xcut_cent = cutcent
      endif else begin
         xcut_cent = 1770. / xbin
      endelse      
      cut = median(obj(xcut_cent-5:xcut_cent+5,*),dim=1)
      z   = lindgen(ysz)
   endif
   ; Select lines
   if keyword_set(manual) then begin
      ; By hand
      window,10,xs=ssz(0),ys=(ssz(1) < 600)
      !p.multi = [0,1,2]
      plot,z,cut,/xstyle,title='Select lines to trace'
      plot,z,cut,/xstyle,yrange=[min(cut),min(cut)+1000]
      print,'Left-click on lines to trace.  Lines should be fairly well spaced.'
      print,'Right-click to exit.'
      getpoints,zline,linecounts,/plotx
      nlines = n_elements(zline)
   endif else begin
      ; Automatically
      zline      = -1
      first_peak = 1
      if blue or newred then begin
         zmin = 0
         zmax = ysz-1
      endif else begin
         zmin = 40    ; Assumes 1x1 binning
         zmax = 2087
      endelse
      for j=zmin+10,zmax-10 do begin
         med_zmin = (0 > (z(j) - 20)) < (max(z) - 41)
         med_zmax = med_zmin + 40
         if (cut(j) eq max(cut(j-5:j+5)) and $
             cut(j) gt median(cut(med_zmin:med_zmax)) + $
                          5*robust_sigma(cut(med_zmin:med_zmax)) and $
             cut(j) lt 60000. and $          ; Avoid saturated lines
             cut(j) ne cut(j-1)) then begin  ; Avoid double-counting
            if first_peak then begin
               zline      = [z(j)]
               first_peak = 0
            endif else begin
               zline      = [zline,z(j)]
            endelse
         endif
      endfor
      nlines = n_elements(zline)
      ; Show located lines
      window,9,xsize=1000,ysize=500
      !p.multi = [0,1,1]
      plot,z,cut,/xs,title='Selected Lines',/ylog,yr=[1e1,7e4],/ys
      for j=0,nlines-1 do oplot,[zline(j),zline(j)],[1e-6,1e6],color=250,$
                                linestyle=1
   endelse

   if blue or newred then begin
      linex = replicate(xcut_cent,nlines)
      liney = zline
   endif else begin
      linex = zline
      liney = replicate(ycut_cent,nlines)
   endelse

endelse

;;; Optional - write line positions to a file.
if keyword_set(outlist) then begin
   if (size(outlist,/tname) eq 'STRING') then out_list = outlist $
      else out_list = 'list.dat'
   openw,list_lun,out_list,/get_lun
   for i=0,nlines-1 do printf,list_lun,linex(i),liney(i)
   close,list_lun
endif

;;; Set range of slit positions over which to trace lines.
if keyword_set(slitrange) then range=slitrange else begin
   if blue then begin
      case xbin of
         1: range = [1550,2850] ; Assume pre/overscan regions present
         2: range = [825,1465]
      endcase
   endif
   if red then begin
      case ysz of
         1000: range = [100,930]   ; Longslit windowing
         2048: range = [605,1410]  ; Full windowing
         else: begin
                  print,'Error - y-size not recognized.'     
                  return
               end
      endcase
   endif
   if newred then begin
      range = [1470,2030] / xbin  ; TEMPORARY - left-hand chip only
   endif
endelse

;;; Optional - Mask off part of the array when tracing lines
if keyword_set(mask) then begin
   if newred then begin
      print,'MASKING NEEDS TO BE IMPLEMENTED FOR NEW RED MOSAIC'
      return
   endif
   mask_ls = where(slit ge min(mask) and slit le max(mask),n_mask)
   if (n_mask ne 0) then begin
      mask_xy = array_indices(slit,mask_ls)
      mask_x  = mask_xy(0,*)
      mask_y  = mask_xy(1,*)   
      if blue then begin
         ; Shift the LHS, since lines are selected on the RHS
         left_xmax  = max(mask_x)
         xshift     = max(mask_x) - min(mask_x)
         yshift     = 0
         left_half  = obj(0:left_xmax,*)
         left_shift = shift(left_half,xshift,yshift)
         trace_obj  = obj
         trace_obj(0:left_xmax,*) = left_shift
         shift_made = 1   
         ; Kludge: Truncate range of slit positions over which to trace,
         ; assuming units are pixels
         range(0) = range(0) + xshift
      endif else begin
         ; Shift the bottom half, since lines are selected on the top
         ; half
         bottom_ymax  = max(mask_y)
         xshift       = 0
         yshift       = max(mask_y) - min(mask_y)
         bottom_half  = obj(*,0:bottom_ymax)
         bottom_shift = shift(bottom_half,xshift,yshift)
         trace_obj    = obj
         trace_obj(*,0:bottom_ymax) = bottom_shift
         shift_made   = 1
         ; Kludge: Truncate range of slit positions over which to trace,
         ; assuming units are pixels
         range(0) = range(0) + yshift   
      endelse
   endif else begin
      trace_obj
      shift_made = 0
   endelse
endif else begin
   trace_obj  = obj
   shift_made = 0
endelse

;;; Trace lines
chan,11
if keyword_set(sky) then begin
   ; Trace sky features using cross-correlation
   if keyword_set(nside) then n_side=nside else n_side=50
   traceline,trace_obj,linex,liney,xtr,ytr,slittr,xdir=xdir,ydir=ydir,$
      nside=n_side,subpix=10,median=21,coord=slit,cbound=range,merit=merit,$
      /noplot
endif else begin
   ; Trace lines using Gaussian fits
   if keyword_set(nside) then n_side=nside else n_side=7
   traceline,trace_obj,linex,liney,xtr,ytr,slittr,xdir=xdir,ydir=ydir,$
      nside=n_side,/bygauss,median=3,coord=slit,cbound=range,merit=merit,$
      /noplot
endelse

if shift_made then begin
   ; Undo temporary shift
   if blue then begin
      shift_ls = where(xtr le left_xmax)
      xtr(shift_ls) = xtr(shift_ls) - xshift
      ytr(shift_ls) = ytr(shift_ls) - yshift
   endif else begin
      shift_ls = where(ytr le bottom_ymax)
      xtr(shift_ls) = xtr(shift_ls) - xshift
      ytr(shift_ls) = ytr(shift_ls) - yshift
   endelse
endif
if blue then begin
   ; Adjust x and y values of pixels on the RH chip to account
   ; for the chip offset and rotation.
   xoffset = 103.192 / xbin
   yoffset = -2.982 / ybin
   xscale  = 0.00162 * (ybin/xbin)
   ; Assume pre-scan region is present
   case xbin of
     1: rh_ls = where(xtr ge 2252,n_rh)
     2: rh_ls = where(xtr ge 1124,n_rh)      
     else : begin
                print,'Error - Not sure how to handle binning.'
                return
            end
   endcase
   if (n_rh ne 0) then begin
      xtr(rh_ls) = xtr(rh_ls) + xoffset + xscale*(ytr(rh_ls)+1)
      ytr(rh_ls) = ytr(rh_ls) + yoffset
   endif
   pixtr = xtr    ; Independent variable
   ztr   = ytr    ; Dependent variable
endif else begin
   if red then begin
      pixtr = ytr
      ztr   = xtr
   endif else begin
      pixtr = xtr
      ztr   = ytr
   endelse
endelse

;;; Fit polynomials.  Reject outliers and iterate until 
;;; fit converges.
if blue then trace_order = 3 else trace_order = 3 ;4
b           = fltarr(trace_order+1,nlines)
berr        = fltarr(trace_order+1,nlines)
zfit        = 0*ytr
nacross     = floor(sqrt(nlines))
ndown       = ceil((1.*nlines)/nacross)
window,11,xs=800,ys=600
!p.multi    = [0,nacross,ndown]
for i=0,nlines-1 do begin
   temp_ls  = where(merit(*,i) ne 0,n_temp)
   temp_pix = pixtr(temp_ls,i)
   temp_z   = ztr(temp_ls,i)
   fit_converged = 0
   old_zfit      = 0
   n_iter        = 0
   good_ls       = indgen(n_temp)
   while not(fit_converged) do begin
      temp_b = poly_fit(temp_pix(good_ls),temp_z(good_ls),trace_order,$
                        yerror=zerr,sigma=temp_berr)
      temp_zfit   = 0*temp_pix
      for j=0,trace_order do temp_zfit = temp_zfit + temp_b(j)*temp_pix^j
      fit_diff = temp_zfit - old_zfit
      if (max(abs(fit_diff)) eq 0) then begin
         fit_converged = 1
      endif else begin
         good_ls  = where(abs(temp_z-temp_zfit) le 3*zerr,n_good)
         old_zfit = temp_zfit
         if (n_good lt trace_order+1) then begin
            ; Flag bad fit
            print,'Bad fit - rejecting.'
            temp_b(*)     = -9999
            temp_berr(*)  = -9999
            fit_converged = 1
         endif
      endelse
      n_iter = n_iter + 1
      if (n_iter ge 10) then fit_converged = 1
   endwhile
   b(*,i) = temp_b
   berr(*,i) = temp_berr
   ; Plot residuals
   temp_zresid = temp_z - temp_zfit
   plot,temp_pix,temp_zresid,xstyle=3,/ystyle,yr=[-1,1],psym=3,$
      title='x = '+strtrim(linex(i),2)+', y = '+strtrim(liney(i),2),$
      charsize=2
endfor

;;; Fit coefficients as a function of intercept
; Reject bad fits
if (blue or newred) then begin
   fitls = where(b(0,*) ne -9999 and b(0,*) ge -500 and b(0,*) le ysz+500)
endif else begin
   fitls = where(b(0,*) ne -9999 and b(0,*) ge -500 and b(0,*) le xsz+500)
endelse
b0 = b(0,fitls)
q  = fltarr(trace_order,4)
window,10,xs=(700 < ssz(0)),ys=(700 < ssz(1))
!p.multi = [0,1,1]
for i=0,trace_order-1 do begin
   print,' '
   print,'Fitting term '+strtrim(i+1,2)
   bi    = b(i+1,fitls)
   berri = berr(i+1,fitls)
   qi    = inter_poly_fit(b0,bi,1,include=inc,measure_errors=berri)
   q(i,0:n_elements(qi)-1) = qi
endfor

;;; Construct the wavelength grid.
print,'Computing wave grid...'
if blue then begin
   x = indgen(xsz)
   y = indgen(ysz)
   wavearr = float(xy2b(x,y,q,/lrisb,xbin=xbin,ybin=ybin))
endif
if red then begin
   x = indgen(xsz)
   y = indgen(ysz)
   wavearr = float(xy2a(x,y,q))
endif
if newred then begin
   x = indgen(xsz)
   y = indgen(ysz)
   wavearr = float(xy2b(x,y,q))
endif

;;; Optional - write to file.
if (size(wavegrid,/tname) eq 'STRING') then $
   writefits,wavegrid,wavearr else wavegrid = wavearr

return
end
