Function im_offset,im1,im2,nx=nx,ny=ny,subx=subx,suby=suby,filter=filter

;+---------------------------------------------------------------------------
;
; IM_OFFSET    3/2007
;
; Find the best integer (x,y) shift that matches im1 to im2 by computing
; the 2D crosscorrelation as a function of lag in x and y.
;
; A shifted version of im1 that matched im2 can be created using IM_SHIFT:
;                 shifted_im1 = im_shift(im1,offset)
; where 'offset' is the shift returned by this program.
;
; INPUTS
;        im1      - 2D array to be shifted
;        im2      - 2D array, same size as im1, to which im1 is to be
;                   matched
;
; KEYWORDS
;        nx       - Integer, number of +/- steps in x over which to
;                   compute crosscorrelation.  Default is 10.
;        ny       - Integer, number of +/- steps in y over which to
;                   compute crosscorrelation.  Default is 10.
;        subx     - 2-element array, [xmin,xmax] of im1 to be used to match
;                   to im2.  Default is to use the central 200 pixels.
;                   A smaller range will decrease the computing time
;                   but may give less accurate results.
;        suby     - 2-element array, [ymin,ymax] of im1 to be used to match
;                   to im2.  Default is to use the central 200 pixels.
;                   A smaller range will decrease the computing time
;                   but may give less accurate results.
;        shiftim  - 2D array, version of im1 shifted to match im2
;        filter   - Integer, box size for median filtering of
;                   images prior to computing crosscorrelation.  Useful
;                   for eliminating bad pixels that may have the same
;                   (x,y) in both images.  Default is no filtering.
;
; OUTPUTS
;        <result> - 2-element array of offsets [delta_x,delta_y] which
;                   gives the best match to im2., i.e.: 
;                          im1(i,j) ~ im2(i+delta_x,j+delta_y).
;                   
; HISTORY
;        Written 3/3/2007 GDB
;----------------------------------------------------------------------------

if (n_params() lt 2) then begin
   print,'CALLING SEQUENCE:'
   print,'   shift = im_offset(im1,im2,nx=nx,ny=ny,subx=subx,suby=suby,'
   print,'                     filter=filter)'
   return,-1
endif

; Get sizes of input arrays
orig_sz1 = size(im1)
orig_sz2 = size(im2)
case orig_sz1(0) of
   2:    begin
            if (orig_sz1(1) ne orig_sz2(1)or $
                orig_sz1(2) ne orig_sz2(2)) then begin
               print,'IM_OFFSET: im1 and im2 must have the same dimensions'
               return,-1
            endif
         end
   else: begin
            print,'IM_OFFSET: im1 and im2 must be 2D'
            return,-1
         end
endcase
orig_xsz = orig_sz1(1)
orig_ysz = orig_sz1(2)

; Set number of steps in x and y over which to compute crosscorrelation
if (keyword_set(nx)) then n_xside = nx else n_xside = 10
if (keyword_set(ny)) then n_yside = ny else n_yside = 10
max_xside = floor((orig_xsz-1)/2.)
max_yside = floor((orig_ysz-1)/2.)
n_xside   = n_xside < max_xside
n_yside   = n_yside < max_yside
n_xstep   = 2*n_xside + 1
n_ystep   = 2*n_yside + 1
xoffset   = lindgen(n_xstep)-((n_xstep-1)/2) 
yoffset   = lindgen(n_ystep)-((n_ystep-1)/2)

; Select region to use for crosscorrelation
if keyword_set(subx) then begin
   x1_1 = min(subx)
   x1_2 = max(subx)
endif else begin
   ; Default is to use (at most) then inner 200 pixels
   x1_1 = 0 > (round(orig_xsz/2.) - 100)
   x1_2 = (orig_xsz-1) < (round(orig_xsz/2.) + 99)
endelse
if keyword_set(suby) then begin
   y1_1 = min(suby)
   y1_2 = max(suby)
endif else begin
   ; Default is to use (at most) then inner 200 pixels
   y1_1 = 0 > (round(orig_ysz/2.) - 100)
   y1_2 = (orig_ysz-1) < (round(orig_ysz/2.) + 99)
endelse


; Filter images prior to computing crosscorr (optional)
if keyword_set(filter) then begin
   filt_im1 = median(im1,filter,/even)
   filt_im2 = median(im2,filter,/even)
endif else begin
   filt_im1 = im1
   filt_im2 = im2
endelse

; Sub-image to be matched to im2
subim1 = filt_im1(x1_1:x1_2,y1_1:y1_2) - $
                  median(filt_im1(x1_1:x1_2,y1_1:y1_2))
tot_subim1_sq = total(subim1^2)

; Crosscorrelation values
cc = fltarr(n_xstep,n_ystep)

; Compute crosscorrelation for each offset
for i=0,n_xstep-1 do begin				 
   for j=0,n_ystep-1 do begin				 
      xoff    = xoffset(i)				         
      yoff    = yoffset(j)				         
      x2_1    = x1_1 + xoff				 
      x2_2    = x1_2 + xoff				 
      y2_1    = y1_1 + yoff				 
      y2_2    = y1_2 + yoff				 
      subim2  = filt_im2(x2_1:x2_2,y2_1:y2_2) - $
                   median(filt_im2(x2_1:x2_2,y2_1:y2_2))
      tot_subim2_sq = total(subim2^2)
      cc(i,j) = total(subim1 * subim2) / sqrt(tot_subim1_sq * tot_subim2_sq)
   endfor						 
endfor

; Determine best offset
max_ind    = where(cc eq max(cc))
offset_ind = array_indices(cc,max_ind)
best_xoff  = xoffset(offset_ind(0))
best_yoff  = yoffset(offset_ind(1))

if (abs(best_xoff) eq n_xside) then $
   print,'IM_OFFSET: Warning - Best x offset is at edge of range.'
if (abs(best_yoff) eq n_yside) then $
   print,'IM_OFFSET: Warning - Best y offset is at edge of range.'

; Return best shift
return,[best_xoff,best_yoff]
end
